/**
 * Copyright (C) 2012 Matteo Tamigi
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, 
 * USA.
*/

#include "PDFSec.hpp"
#include "md5.h"
#include "rc4.h"

static void
printHexString(vector<uint8_t> str) {
  for(int i=0;i<str.size();i++)
    printf("%x ",str[i]);
  printf("\n");
}

static void
printHexString(const uint8_t * str, int size) {
  for(int i=0;i<size;i++)
    printf("%x ",str[i]);
  printf("\n");
}

//Costanti di cifratura per inizializzare i vari algoritmi di cifratura
const uint8_t pad[] = {
  0x28,0xBF,0x4E,0x5E,0x4E,0x75,0x8A,0x41,
  0x64,0x00,0x4E,0x56,0xFF,0xFA,0x01,0x08,
  0x2E,0x2E,0x00,0xB6,0xD0,0x68,0x3E,0x80,
  0x2F,0x0C,0xA9,0xFE,0x64,0x53,0x69,0x7A
};


//Funzioni per la cifratura/decifratura delle varie versioni

bool PDFSec :: isUserPasswordRev2 () {
    md5(key,sizeKey,digest);                                  //Cifro la key con l'md5
    //Eseguo un test sui primi 4 caratteri 
    Rc4_40( u_string , 3 , digest , tmp );    //Cifro la u_str con le 5 cifre prese prima
    //Devo ottenere in dec il pad
    for(unsigned i = 0; i < 3; i++)
        if(tmp[i] != pad[i])
            return false;
    //------------------------------------------------------------------------//
    //Test superato, proseguo con quello completo (prendendo solo i primi 5 caratteri dell'md5)
    Rc4_40( u_string , 32 , digest , tmp);    //Cifro la u_str con le 5 cifre prese prima
    //Devo ottenere in dec il pad
    for(unsigned i = 0; i < 32; i++)
        if(tmp[i] != pad[i])
            return false;
    //Superato il test, la chiave  quella giusta
    return true;
}

bool PDFSec :: isUserPasswordRev3 () {
    //Cifro la key per ottenere cifr
    md5(key,sizeKey,digest);                //Cifro la key con l'md5
    md5_50(digest);                           //Cifro in md5 altre 50 volte (nel dubbio :D)
    //Eseguo un test sulle prime 3 cifre (per evitare di farlo tutto), se lo passa
    //eseguo il test completo
    for(int i = 0; i < 3; i++)
        tmp[i] = infoPdf->u_string[i];
    //Cifro con l'RC4 su 3 cifre
    for(int i = 19; i >= 0; --i) {
        for(unsigned j = 0; j < length; ++j)
            tmpkey[j] = digest[j] ^ i;
        Rc4_128(tmp,3,tmpkey,tmp);
    }
    //Confronto il risultato atteso cifr con la user_pswd cifrata
    for(unsigned i = 0; i < 3; i++)
        if(tmp[i] != cifr[i])
            return false;
    //------------------------------------------------------------------------//
    //Il test precedente ha avuto buon esito, ora effettuo la verifica totale
    for(int i = 0; i < 16; i++)
        tmp[i] = infoPdf->u_string[i];
    //Cifro con l'RC4 su 3 cifre
    for(int i = 19; i >= 0; --i) {
        for(unsigned j = 0; j < length; ++j)
            tmpkey[j] = digest[j] ^ i;
        Rc4_128(tmp,16,tmpkey,tmp);
    }
    //Confronto il risultato atteso cifr con la user_pswd cifrata
    for(unsigned i = 0; i < 16; i++)
        if(tmp[i] != cifr[i])
            return false;
    //Test finale passato!
    return true;
}


//Funzioni della classe

void PDFSec :: initCracking ( const string usr_pswd = "" ) {
    vector<uint8_t> keyy;
    //Stabilisco la versione della cifratura ed effettuo le operazioni preliminari:
    //-Verifico che i parametri in ingresso siano corretti
    //-Se esiste la user password, verifico che sia quella corretta
    //-Genero il dizionario per le password di bruteforcing
    //------------------------------------------------------------------------//
    
    /*--------------------------------------------------------------------------
    Standard di cifratura PDF:
    V: 1    R: 1 - 3    RC4 a 40-bit con parole da 32 caratteri Latin-1
            R: 4        RC4 a 128-bit con parole da 32 caratteri Latin-1
            R: 5        come la R4 solo diverso metodo di cifratura (NO SPECIFICHE)
            R: 6 - 7    AES-128 con parole da 32 caratteri Latin-1 (NO SPECIFICHE)
            R: 7 ext 3  AES-256 con parole da 127 UTF-8 bytes
                 ext 8  come ext 3 ma diverso metodo di cifratura
    --------------------------------------------------------------------------*/
    
    //Riempio la chiave da cifrare
    // password_padded + o_string + permissions + fileID [ (se la revision  > 3) + encryptMetaData]
    
    //Se c' la user password la inserisco e la utilizzo
    if(usr_pswd != "") {
        userPswd = usr_pswd;
        isUserPswd = true;
        if( 32 < userPswd.length())
            userPswd.erase(userPswd.begin()+32,userPswd.end());    //Se la password  superiore ai 32 caratteri, la taglio
        //Riempio la chiave di cifratura con la user_pswd
        for(unsigned i = 0; i < userPswd.length(); i++)
            keyy.push_back((uint8_t)userPswd[i]);
    } else 
        isUserPswd = false;
    
    for(unsigned i = 0; i < 32-userPswd.length(); i++)
        keyy.push_back(pad[i]);
    for(unsigned i = 0; i < 32; i++)
        keyy.push_back(infoPdf->o_string[i]);
    keyy.push_back(infoPdf->permissions & 0xff);
    keyy.push_back((infoPdf->permissions >> 8) & 0xff);
    keyy.push_back((infoPdf->permissions >> 16) & 0xff);
    keyy.push_back((infoPdf->permissions >> 24) & 0xff);
    for(unsigned i = 0; i < infoPdf->fileID.size(); i++)
        keyy.push_back(infoPdf->fileID[i]);
    if(infoPdf->revision > 3 && !infoPdf->encryptMetaData) {
        keyy.push_back(0xff);
        keyy.push_back(0xff);
        keyy.push_back(0xff);
        keyy.push_back(0xff);
    }
    
    //Copio i risultati sul vettore key (che utilizzer per il cracking)
    key = new uint8_t[keyy.size()];
    sizeKey = keyy.size();
    for(unsigned i = 0; i < keyy.size(); i++)
        key[i] = keyy[i];
    u_string = new uint8_t[infoPdf->u_string.size()];
    for(unsigned i = 0; i < infoPdf->u_string.size(); i++)
        u_string[i] = infoPdf->u_string[i];
    
    //Il risultato dell'md5
    digest = new uint8_t[16];
    
    switch( infoPdf->revision ) {
        case 2:
            tmp = new uint8_t[32];
            break;
        case 3:
            //Genero la chiave di test (diversa dal pad)
            //md5 ( pad + fileID )
            uint8_t * tempCifr;
            cifr = new uint8_t[32 + infoPdf->fileID.size()];
            for(unsigned i = 0; i < 32; i++)
                cifr[i] = pad[i];
            for(unsigned i = 0; i < infoPdf->fileID.size(); i++)
                cifr[32 + i] = infoPdf->fileID[i];
            md5(cifr,32 + infoPdf->fileID.size(),digest);
            delete [] cifr;
            cifr = new uint8_t[16];
            for(unsigned i = 0; i < 16; i++)
                cifr[i] = digest[i];
            //Genero le variabili di lavoro
            length = infoPdf->length/8;
            tmp = new uint8_t[32];
            tmpkey = new uint8_t[length];
            break;
    }
}

string PDFSec :: crackUserPassword () {
    vector<uint8_t> pswd_padd;
    char * parola;
    //Verifico che sia stato inizializzato l'ambiente di cracking
    if( key == NULL ) {
        cerr << "WARNING: Cracking workspace isn't initialized! Initializing..." << endl;
        initCracking();
    }
    
    //Verifico se la user_password sia quella nulla (per sicurezza)
    switch ( infoPdf->revision ) {
        case 2:
            if( !isUserPswd && isUserPasswordRev2() )
                return "";
            SetCharset(STDCHARSET);
            SetState(1,5,NULL);
            //Faccio partire il ciclo di cracking
            while ( !isUserPasswordRev2() ) {
                SetNextPswd();
            }
            return GetPswd();
            break;
        case 3:
            if( !isUserPswd && isUserPasswordRev3() )
                return "";
            SetCharset(STDCHARSET);
            SetState(1,5,NULL);
            //Faccio partire il ciclo di cracking
            while ( !isUserPasswordRev3() ) {
                SetNextPswd();
            }
            return GetPswd();
            break; 
    }
}

void PDFSec :: SetCharset ( const string chs ) {
    if(charSet)
        delete [] charSet;
    lenSet = chs.length();
    charSet = new uint8_t[lenSet];
    for(unsigned i = 0; i < lenSet; i++)
        charSet[i] = (uint8_t)chs[i];
}

void PDFSec :: SetState ( const unsigned min = 1, const unsigned max = 5, unsigned * st = NULL ) {
    //Verifico gli estremi di generazione, se non son validi
    //imposto gli estremi da 1 a 5 di default
    if( min > max && 1 < max && 1 < min ) { min_c = 1; max_c = 5; }
    else { max_c = max; min_c = min; }
    //Genero il vettore degli stati e lo carico (o lo azzero)
    if(state)
        delete [] state;
    state = new unsigned[max_c];
    if( st ) {
        //Copio lo stato
        for(unsigned i = 0; i < max_c; i++)
            state[i] = st[i];
        //Determino la lunghezza
        for(unsigned i = 0; i < max_c; i++)
            if(state[i] == -1)
                lenState = i;
    }
    else {
        for(unsigned i = 0; i < min_c; i++)
            state[i] = 0;
        for(unsigned i = min_c; i < max_c; i++)
            state[i] = -1;
        lenState = min_c;
        
    }
    //Imposto la char
    unsigned i;
    for(unsigned i = 0; i < lenState; i++)
        key[i] = charSet[state[i]];
    for(unsigned i = 0 ; i < 32 - lenState; i++)
        key[lenState + i] = pad[i];
    EOS = false;
}

//Questa funzione lavora direttamente sul vettore key (per risparimiare lavoro)
void PDFSec :: SetNextPswd () {
    //Incremento la combinazione di 1
    register unsigned i = 0;
    while ( i < lenState ) {
        if( ++state[i] < lenSet ) {
            key[i] = charSet[state[i++]];
            break;
        }
        else {
            //Ciclo di password finito
            if( i == max_c ) {
                EOS = true;
                break;
            }
            //Sono all'ultima combinazione con i cifre
            state[i] = 0;
            key[i++] = charSet[0];
            if( i == lenState ) {
                state[lenState] = 0;
                key[lenState] = charSet[0];
                lenState++;
                for(unsigned j = 0; j < 32 - lenState; j++)
                    key[lenState + j] = pad[j];
                break;
            }
            
        }
    }
}

string PDFSec :: GetPswd () {
    string str;
    for(unsigned i = 0; i < lenState; i++)
        str.push_back( charSet[state[i]] );
    return str;
}
