/**
 * Copyright (C) 2012 Matteo Tamigi
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, 
 * USA.
*/

#include "PDFSec.hpp"
#include <stdlib.h>

#define BUFFSIZE 256

//Funzioni su file e stringhe
uint8_t hexToInt(const int b) {
    if(b >= '0' && b <= '9')
        return b-'0';
    else if(b >= 'a' && b <= 'f')
        return b-'a'+10;
    else if(b >= 'A' && b <= 'F')
        return b-'A'+10;
    else
        return 0;
}

//Parte della classe PDFSec che si occupa della lettura del file PDF e della
//ricerca delle informazioni contenute al suo interno. l'intera parte  privata
//e resa trasparente dalla funzione LoadPdfInfos ();

int PDFSec :: readPdfFile ( ) {
    //DEBUG
    //------------------------------------------------------------------------//
    if(verbose)
        cout << "[readFile] Entered in th function..." << endl; 
    //------------------------------------------------------------------------//
    int vers, subvers, ret;
    char * buffer = new char[BUFFSIZE];
    string buf;
    int pos_t;
    //Recupero la prima linea del file
    is.getline(buffer, BUFFSIZE);
    buf = buffer;
    //Verifico che la prima linea del file sia l'intestazione di un PDF e ne leggo
    //la versione
    if( -1 != buf.find("%PDF-") ) {
        //DEBUG
        //--------------------------------------------------------------------//
        if(verbose)
            cout << "[readFile] Found '%PDF-'" << endl; 
        //--------------------------------------------------------------------//
        //Estraggo la versione nel formato "vers.subvers", esempio: "%PDF-1.4"
        vers = atoi(&buffer[5]);
        subvers = buf.find_first_of(".",6);
        subvers = atoi(&buffer[subvers+1]);
        
        //DEBUG
        //--------------------------------------------------------------------//
        if(verbose)
            cout << "[readFile] Version: " << vers << " Subversion: " << subvers << endl; 
        //--------------------------------------------------------------------//
        
        if( 0 <= vers && 0 <= subvers ) {
            infoPdf->version_major = vers;
            infoPdf->version_minor = subvers;
        }
        else
            return ERRVERS; //Versioni errate, probabilmente non  un PDF
    }
    else
        return ERRVERS;     //Il file non  un PDF
    
    //Mi sposto verso la fine del file ed eseguo la ricerca della sezione
    // "trailer\n<</Encrypt ... [<file_id><id>]>>"
    
    //DEBUG
    //------------------------------------------------------------------------//
    if(verbose)
        cout << "[readFile] Search trailer..." << endl; 
    //------------------------------------------------------------------------//
    
    is.seekg(-1024,ios::end);
    pos_t = findTrailer ();
    
    //DEBUG
    //------------------------------------------------------------------------//
    if(verbose)
        cout << "[readFile] First result (from eof-1024): " << pos_t << endl; 
    //------------------------------------------------------------------------//
    
    //Se non ho trovato il trailer, parto a cercarlo dall'inizio, per esser
    //sicuro di trovarlo
    if( pos_t < 0 ) {
        is.clear();
        is.seekg(0,ios::beg);
        pos_t = findTrailer ();
        
        //DEBUG
        //------------------------------------------------------------------------//
        if(verbose)
            cout << "[readFile] Second result: " << pos_t << endl; 
        //------------------------------------------------------------------------//
    
    }
    //Nel caso ci sia un errore, lo ritorno (o non c' il trailer oppure manca qualcosa)
    if( pos_t < 0 )
        return pos_t;
    ret = getEncInfo ( pos_t );     //findEncryptObject(file,e_pos);
    
    //DEBUG
    //------------------------------------------------------------------------//
    if(verbose && ret)
        cout << "[readFile] EncInfo OK" << endl;
    else if( verbose && !ret )
        cout << "[readFile] EncInfo ERR" << endl;
    //------------------------------------------------------------------------//
    
    
    if(!ret)
        return ERRNOENCINFO;
    
    return 0;
}



//  findTrailer
//  cerco la seguente stringa:
//  "trailer<</Encrypt ... [<file_id><id>]>>"
int PDFSec :: findTrailer ( ) {
    int ch, pos_i, e_pos;
    bool encrypt = false;
    bool id = false;
    vector<uint8_t> str;
    string buf;
    
    //Scarto la fine della linea in corso (il trailer inizia con una nuova riga)
    //ATTENZIONE!!! NEL CASO IL FINE STRINGA SIA SOLTANTO 0x0A OPPURE 0x0D e non
    //TUTTI E DUE, L'ALGORITMO DA UN PROBLEMA
    //------------------------------------------------------------------------//
    getline(is,buf);
    //------------------------------------------------------------------------//
    while ( !is.eof() ) {
        getline(is,buf);
        if( -1 != buf.find("trailer",0) ) {
            //Ho trovato il trailer, la prossima riga (dovrebbe)  la stringa 
            //che contiene anche le informazioni di criptazione e inizia con "<<"
            //ATTENZIONE, ZONA CRUCIALE DA SCRIVERE CON MAGGIORI CONTROLLI!!!
            //----------------------------------------------------------------//
            do {
                getline(is,buf);
            } while( -1 == buf.find("<<",0) );
            //----------------------------------------------------------------//
            //Sono sulla stringa corretta, ricerco le info:
            //"/Encrypt 36 0 R"
            pos_i = buf.find("/Encrypt",2);
            if( -1 != pos_i ) {
                //recupero il primo numero che trovo ( scritto malissimo, ma funziona)
                e_pos = atoi ( buf.substr( pos_i+8 , buf.find_first_of('/', pos_i + 8) - pos_i - 8 ).c_str() );
                if ( e_pos >= 0 )
                    encrypt = true;
            }
            //"/ID [<...><...>]"
            pos_i = buf.find("/ID",2);
            if( -1 != pos_i ) {
                //estraggo la stringa racchiusa tra parentesi quadre
                int estr_inf = buf.find_first_of('[' , pos_i+3) + 1,
                    estr_sup = buf.find_first_of(']' , pos_i+64);
                str = parseID( buf.substr( estr_inf , estr_sup - estr_inf ) );
                if(!str.empty())
    		        id = true;
            }
            
            if(encrypt && id) {
                infoPdf->fileID = str;
                //infoPdf->fileIDLen = str->len;
                //delete str;
                return e_pos;
            }
        }
    }
    if(!encrypt && id)
        return ERRNOENCINFO;
    else if(!id && encrypt)
        return ERRNOFILEID;
    else 
        return ERRNOINFOTRA;
}

//Recupero l'ID del file "<...><...>" di 32 byte contenuto tra le prima parentesi
//acute
vector<uint8_t> PDFSec :: parseID ( string str ) {
    vector<uint8_t> ret;
    uint8_t buf[BUFFSIZE];
    int len = 0, i = 0;
    //Estraggo i primi 32 byte
    while(str[i] != '>' && len < BUFFSIZE && i < str.length() ) {
        if((str[i] >= '0' && str[i] <= '9') || (str[i] >= 'a' && str[i] <= 'f') || (str[i] >= 'A' && str[i] <= 'F')) {
	        buf[len++] = str[i];
        }
        i++;
    }
    //Li converto in esadecimale
    //ret->content = new uint8_t[len/2];
    //ret->len = (len/2);
    
    for(unsigned j = 0, h = 0; j<len; j += 2) {
        ret.push_back(hexToInt(buf[j]) * 16);
        ret[h] += hexToInt(buf[j+1]);
        h++;
    }

    return ret;
}


bool PDFSec :: getEncInfo ( const int e_pos ) {
    //DEBUG
    //------------------------------------------------------------------------//
    if(verbose)
        cout << "-[getEncInfo] Entered in the function..." << endl; 
    //------------------------------------------------------------------------//
    
    //Converto l'intero in una stringa
    char * pos = new char[BUFFSIZE];
    char * buffer = new char[BUFFSIZE];
    string buf, temp;
    sprintf(pos,"%d",e_pos);
    int ch;
    //Riavvolgo il file per la ricerca
    is.clear();
    is.seekg(0,ios::beg);
    //Ricerco la posizione presente in e_pos
    //esempio di posizione e stringa: "8031 0 obj<<...>>"
    while ( !is.eof() ) {
        buf.clear();
        getline(is,buf);
        if( -1 != (ch = buf.find(pos,0)) ) {
            //DEBUG
            //------------------------------------------------------------------------//
            if(verbose)
                cout << "-[getEncInfo] Find pos_t: " << pos << " File pos: " << is.tellg() << " Col: "<< ch << endl;
            //------------------------------------------------------------------------//
            //non necessariamente (ho parsato solo 3 file) il "<<"  sulla stessa
            //riga dell'e_pos e della stringa "obj" ma l'e_pos  all'inizio della riga
            //se identifichiamo il fine-riga come '0x0a', '0x0d' insieme o in alternativa
            
            
            //Verifico che i caratteri precedenti siano un 'a capo'
            if( 0 == ch || ( 0x0a == buf[ch-1] || 0x0d == buf[ch-1] ) ) {
                //Taglio tutta la stringa precedente a e_pos
                buf.erase(0,ch);
                temp = buf;
            
                //Ho identificato la riga che comincia per pos, carico tutto in
                //una stringa fino a "endobj"
                while( -1 == temp.find("endobj",0) ) {
                    getline(is,temp);
                    buf += temp;
                }
                //nel caso che abbia recuperato troppe righe (generatore PDF del cazzo)
                //taglio la stringa dal punto in cui comincia pos, fino a "endobj"
                buf.erase( buf.find("endobj",0) + 6 );
                //Faccio il parsing della stringa
                //"e_pos 0 obj<< ... >>endobj"
                
                //DEBUG
                //------------------------------------------------------------------------//
                if(verbose)
                    cout << "-[getEncInfo] Find pos_t string correctly: " << buf.substr(0,35) << endl;
                //------------------------------------------------------------------------//
                return parseEncObject( buf );
            }
        }
    }
    return false;
}

bool PDFSec :: parseEncObject ( const string obj ) {
    int ch, i;
    bool fe = false,
         ff = false,
         fl = false,
         fo = false,
         fp = false,
         fr = false,
         fu = false,
         fv = false;
    vector<uint8_t> str;
    //I parametri da recuperare sono:
    //EncryptMetadata, Filter, Length (lunghezza sha-1)
    //O (owner_str), P (permessi), U (user_str), V (versione), R (revisione)
    
    //ATTENZIONE!!! DEBOLEZZA IN QUANTO NON CONTO SPAZI ULTERIORI A 1
    //------------------------------------------------------------------------//
    //EncryptMetadata
    if( -1 != (ch = obj.find("/EncryptMetadata",0)) ) {
        //Cerco il "false" esattamente uno spazio dopo
        if( -1 != obj.find("false", ch+17,10) ) {
            fe = true;
        }
    }
    //Filter
    //Esempio: "/Filter/Standard"
    if( -1 != (ch = obj.find("/Filter",0)) ) {
        ch = obj.find_first_of('/',ch+1);
        i = obj.find_first_of('/',ch+1) - ch - 1;
        if ( ch < ch + i ) {
            infoPdf->s_handler = obj.substr(ch+1,i);
            ff = true;
        }
    }
    //Lenght
    if ( -1 != (ch = obj.find("/Length",0)) ) {
        i = atoi ( obj.substr(ch+8,5).c_str() );
        if ( 0 < i ) {
            infoPdf->length = i;
            fl = true;
        }
    }
    //O (owner_str)
    if( -1 != (ch = obj.find("/O",0)) ) {
        //Riduco la stringa a "(...)" contenente la owner_str alla funzione per il parsing
        int start = obj.find_first_of('(',ch);
        int stop;
        for(int j = start; j < obj.length(); j++){
            if ( obj[j] == ')' && obj[j-1] != '\\' ) {
                stop = j+1;
                break;
            }
        }
        //converto la stringa contenuta tra parentesi in una stringa di byte
        str = stringToByte( obj.substr(start+1,stop-start-1) );
        if( !str.empty() ){
            if(str.size() != 32) {
                cerr << "WARNING: O-String != 32 Bytes: " << str.size() << endl;
                wrn = WRNOSTR;
            }
            infoPdf->o_string = str;
            fo = true;
        }
    }
    //U (user_str)
    if( -1 != (ch = obj.find("/U",0)) ) {
        //Riduco la stringa a "(...)" contenente la owner_str alla funzione per il parsing
        int start = obj.find_first_of('(',ch);
        int stop;
        for(int j = start; j < obj.length(); j++){
            if ( obj[j] == ')' && obj[j-1] != '\\' ) {
                stop = j+1;
                break;
            }
        }
        //converto la stringa contenuta tra parentesi in una stringa di byte
        str = stringToByte( obj.substr(start+1,stop-start-1) );
        if( !str.empty() ){
            if(str.size() != 32) {
                cerr << "WARNING: U-String != 32 Bytes: " << str.size() << endl;
                wrn = WRNUSTR;
            }
            infoPdf->u_string = str;
            fu = true;
        }
    }
    //P (permessi)
    if( -1 != (ch = obj.find("/P",0)) ) {
        //Acquisisco 11 caratteri massimi in quanto il maggior numero negativo 
        //composto da 10 numeri e il -
        infoPdf->permissions = atoi ( obj.substr(ch+3,11).c_str() );
        fp = true;
    }
    //V (versione)
    if( -1 != (ch = obj.find("/V",0)) ) {
        i = atoi ( obj.substr(ch+3,4).c_str() );
        if ( 0 < i ) {
            infoPdf->version = i;
            fv = true;
        }
    }
    //R (revision)
    if( -1 != (ch = obj.find("/R",0)) ) {
        i = atoi ( obj.substr(ch+3,4).c_str() );
        if ( 0 < i ) {
            infoPdf->revision = i;
            fr = true;
        }
    }
    //------------------------------------------------------------------------//
    if(!fe)
        infoPdf->encryptMetaData = true;
    if(!fl)
        infoPdf->length = 40;
    if(!fv)
        infoPdf->version = 0;
        
    if( infoPdf->s_handler != "Standard" )
        return true;
        
    return ff && fo && fp && fr && fu;
}


vector<uint8_t> PDFSec :: stringToByte ( const string s ) {
    unsigned i, j, l;
    uint8_t b, d;
    uint8_t *tmp = new uint8_t[s.length()];
    vector<uint8_t> ret;
    
    for(i=0, l=0; i<s.length(); i++, l++) {
        b = s[i];
        if(b == '\\') {
            //We have reached a special character or the beginning of a octal 
            //up to three digit number and should skip the initial backslash
            switch( s[++i] ) {
                case 'n':
	                b = 0x0a;
	                break;
                case 'r':
                	b = 0x0d;
                	break;
                case 't':
                    b = 0x09;
                	break;
                case 'b':
                	b = 0x08;
                	break;
                case 'f':
                    b = 0x0c;
                	break;
                case '(':
                	b = '(';
                	break;
                case ')':
                	b = ')';
                	break;
                case '\\':
                	b = '\\';
                	break;
                default:
                	if(s[i] >= '0' && s[i] < '8') {
                	   d = 0;
                	   for(j=0; i < s.length() && j < 3 &&
                	     s[i] >= '0' && s[i] < '8' &&
                		 (d*8)+(s[i]-'0') < 256; j++, i++) {
                	       d *= 8;
                	       d += (s[i]-'0');
                	   }
                	   //We need to step back one step if we reached the end of string
                	   //or the end of digits (like for example \0000)
                	   if(i < s.length() || j < 3) {i--;}
                       b = d;
                    }
            }
        }
        tmp[l] = b;
    }
    
    for(unsigned i = 0; i < l-1; i++)
        ret.push_back(tmp[i]);

    return ret;
}

